/* --------------------------------------------------------------  */
/* (C)Copyright  2007,                                             */
/* Andreas Stiller                                                 */
/* Redaktion c't                                                   */
/* Heise Zeitschrifen Verlag GmbH & Co Kg                          */
/*                                                                 */
/* All Rights Reserved.                                            */
/* --------------------------------------------------------------  */

//#define DEBUG

#include <termios.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <libspe2.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sched.h>
#include <linux/kd.h>
#include <linux/fb.h>
#include "spu/spe.h"
#include "colorscheme.h"

extern spe_program_handle_t ctapfel_spu;

#define INTERLEAVE
//Interleave: die Zeilen werden nacheinander auf die SPEs verteilt
//            sonst blockweise

#define WITHGRAPHICS
// Mit Grafikausgabe auf den PS3 Framebuffer

#ifdef WITHGRAPHICS  
 #define USE2BUFFER
//USE2BUFFER verwendet zwei Buffer für Grafikausgabe 
 #include "ps3fb.h"
struct ps3fb_ioctl_res res; 
  
#endif

#define SPU_THREADS 6
#define BPP 4 
#define WIDTH 640
#define LINELENGTH 1024 
#define HEIGHT 480
#define REPEATLOOP 0

const int Depth=2000;
const int Repmax=200;
const double g=2.0f;
const int frameoff=0;
double rz=0.442733f;
double iz=0.378092f;
double dr=4.0f/WIDTH; 
double di=4.0f/WIDTH;
double hopping=0.000f;

double r0,i0;

struct timeval t0,t1,dt; 
struct timezone tz;
double ta,te;
unsigned int buffer1[HEIGHT][LINELENGTH] __attribute__ ((aligned (128)));
#ifdef USE2BUFFER
unsigned int buffer2[HEIGHT][LINELENGTH] __attribute__ ((aligned (128)));
#endif


QRgb rainbow[512];
u64 gsum;

double seconds(void) {
    struct timeval t;
    gettimeofday(&t,&tz);
    return (double) t.tv_sec+ ((double)t.tv_usec/1e6); 
}
 
int kbhit(void) { 
        struct termios term, oterm; 
        int fd = 0; 
        int c = 0; 
        tcgetattr(fd, &oterm); 
        memcpy(&term, &oterm, sizeof(term)); 
        term.c_lflag = term.c_lflag & (!ICANON); 
        term.c_cc[VMIN] = 0; 
        term.c_cc[VTIME] = 1; 
        tcsetattr(fd, TCSANOW, &term); 
        c = getchar(); 
        tcsetattr(fd, TCSANOW, &oterm); 
        if (c != -1) 
        ungetc(c, stdin); 
        return ((c != -1) ? 1 : 0); 
}


#ifdef WITHGRAPHICS

void enableCursor(int enable)
{
   int fd = open("/dev/console", O_NONBLOCK);
   if (fd >= 0) {
      ioctl(fd, KDSETMODE, enable ? KD_TEXT : KD_GRAPHICS);
      close(fd);
   }
}


void draw1 (uint32_t* fb){
  int x, y, y0,c ;
  y0=frameoff+HEIGHT*res.xres;
  for (y = 0; y < HEIGHT; y++) {
      for (x = 0; x < WIDTH; x++) {
                 c=buffer1[y][ x];
                 gsum+=c;
                 fb[y0+x]= (c==Depth) ? 0 : rainbow[c & 0x1ff];
      }
    y0 -=res.xres;
   }
}

 #ifdef USE2BUFFER
void draw2 (uint32_t* fb){
  int x, y, y0,c;
  y0=frameoff+HEIGHT*res.xres;
  for (y = 0; y < HEIGHT; y++) {
      for (x = 0; x < WIDTH; x++) {
                 c=buffer2[y][ x];
                 gsum +=c;
                 fb[y0+x]= (c==Depth) ? 0 : rainbow[c & 0x1ff];
      }
    y0 -=res.xres;
   }
} 

 #endif
#endif

void additersum (void)
{
 int i,j;
		for (i=0; i<HEIGHT; i++) 
			for (j=0; j < WIDTH; j++) 
				 gsum +=buffer1[i][j];
}

/*++ neu mit LIbspe2 ++++*/
struct thread_args {
        struct spe_context * ctx; 
        void * argp;
        void * envp;
};


__attribute__((noreturn)) void * spe_thread(void * arg)
{
        int flags = 0;
        unsigned int entry = SPE_DEFAULT_ENTRY;
        int rc;
        struct thread_args * arg_ptr;

        arg_ptr = (struct thread_args *) arg;

#if 0
	if (spe_program_load(arg_ptr->ctx, &simple_spu)) {
                perror("spe_program_load");
                pthread_exit(NULL);
        }
#endif
        rc = spe_context_run(arg_ptr->ctx, &entry, flags, arg_ptr->argp, arg_ptr->envp, NULL);
        if (rc < 0)
                perror("spe_context_run");

        pthread_exit(NULL);
}
/*!!!!!!*/


int main(void)
{
	
/*+*/ 	pthread_t pts[SPU_THREADS];
/*+*/   spe_context_ptr_t ctx[SPU_THREADS];
/*+*/	struct thread_args t_args[SPU_THREADS];
/*+*/   int flags = 0;	
/*+*/

	struct dma_packet cb[SPU_THREADS];
	int i;  
        int spes; 

	int startline,linesperSPU;	
	unsigned long affinity_mask=0xffffffff;  
	int fd; 
	int length;
#ifdef WITHGRAPHICS
	void *addr=NULL;
        uint32_t frame = 0;
        
#endif
	
        int fbok;
	int rep;

        double drx,dfak;
        double gits;
      
        int rc;
        int z;
        dfak=pow(2.0,1.0/12.0);
	printf ("c't-Apfelmännchen für Cell V1.2 mit SDK 2.1, Andreas Stiller, Port für LibSPE2: Gerhard Stenzel\n");
        int spe_count=spe_cpu_info_get (SPE_COUNT_PHYSICAL_SPES,-1);
        printf ("found %d SPEs\n",spe_count); 

	/* Create several SPE-Context to execute 'ctapfel_spu'.
	*/
	spes=spe_count; 
        if (spes > HEIGHT) spes=HEIGHT; // nicht genug Arbeit für alle SPEs
 

/*+*/	for(i=0; i<spes; i++) {
/*+*/		if (!(ctx[i] = spe_context_create(flags, NULL))) {
/*+*/			perror("spe_create_context");
/*+*/			return -2;
/*+*/		}
/*+*/#if 1	      
/*+*/		if (spe_program_load(ctx[i], &ctapfel_spu)) {
/*+*/			perror("spe_program_load");
/*+*/			return -3;
/*+*/		}
/*+*/#endif
        linesperSPU=HEIGHT/spes;            // Zeilen pro SPU (abgerundet)
        if (i < HEIGHT%spes) linesperSPU++; // Den Rest verteilen   
     

        #ifdef INTERLEAVE
      
               cb[i].targetoffs=LINELENGTH*spes;
        #else 
   
	       cb[i].targetoffs=LINELENGTH;
        #endif	
     
	       cb[i].g=g;
	       cb[i].len=WIDTH; 
	       cb[i].depth=Depth; 
     	       cb[i].verbose=0; 
               cb[i].linecnt=linesperSPU;
	}
		
	fd=0;
	fbok=0; 
        length=0;
#ifdef WITHGRAPHICS
	//   enableCursor(0);

	if ((fd = open("/dev/fb0", O_RDWR)) < 0) {
		printf("error open:%d\n", fd);
		fbok=0;
	} 

	if (fd) {
           if (ioctl(fd, PS3FB_IOCTL_SCREENINFO, (unsigned long)&res) < 0) {
		printf("PS3FB_IOCTL_SCREENINFO failed\n");
		fbok=0;
              }
           else fbok=1;
          }



	printf("xres: %d, yres: %d, xoff: %d, yoff: %d, num_frames: %d\n",
		res.xres, res.yres, res.xoff, res.yoff, res.num_frames);
	length = res.xres * res.yres * BPP * res.num_frames -
		res.xres * res.yoff * BPP * 2; /* top and bottom margins */


	if (fbok) { 
		addr = mmap(NULL, length, PROT_WRITE, MAP_SHARED, fd, 0);
		if (!addr) {
			printf("mmap failed\n");
			fbok=0;
			// stop flipping in kernel thread with vsync
			ioctl(fd, PS3FB_IOCTL_ON, 0);

		}  
#ifdef DEBUG     
       else printf ("mmap ok addr=%x \n",(unsigned int) addr);    
#endif        
	}
#endif
       
        z=100;
        do {
	gsum=0;
        drx=dr;
        ta=seconds();
        SetColorScheme(rainbow,(10*z)%512,512); // Credit to Trolltech (QT) 
           
	for (rep=0; rep < Repmax; rep++) 
	{
		r0=rz-drx*WIDTH/2; 
		i0=iz-di*HEIGHT/2;
                startline=0;
		
		for(i=0; i<spes; i++) {
                  cb[i].r  = r0;
		  cb[i].i  = i0+startline*di;
                  cb[i].dr = drx;
	          #ifdef INTERLEAVE
                    cb[i].di= spes*drx;
                  #else 
                    cb[i].di= drx;
                  #endif	
               
			
 
#ifdef USE2BUFFER
                        if (rep & 1) cb[i].targetaddr.ull = (u64) ((u32) &buffer2[startline][0]);
                        else cb[i].targetaddr.ull = (u64) ((u32) &buffer1[startline][0]);
#else
			cb[i].targetaddr.ull = (u64) ((u32) &buffer1[startline][0]);
#endif

#ifdef DEBUG
 printf ("rep %d, spu:%d ab Line %d - %d Targetoffs: %d\n",rep, i,startline,cb[i].linecnt,cb[i].targetoffs);
#endif 
                         
		        #ifdef INTERLEAVE
               		 startline++;
         	        #else 
                         startline+=cb[i].linecnt;
         	        #endif	
   
/*+*/			t_args[i].ctx = ctx[i];
/*+*/			t_args[i].argp = (u64 *)&cb[i];
/*+*/			rc = pthread_create(&pts[i], NULL, &spe_thread, &t_args[i]);
/*+*/			if (rc) {
/*+*/				fprintf(stderr, "Failed pthread_create(rc=%d, mask=%lx errno=%d)\n", 
/*+*/					(int) rc, affinity_mask, errno);

				exit(1);
			}

		}
#ifdef WITHGRAPHICS
 #ifdef USE2BUFFER	

                if (fbok && rep > 0) {
                  if (rep & 1) draw1 ((uint32_t*) (u32) addr + frame * res.xres *  res.yres*4);
                  else  draw2 ((uint32_t*) (u32) addr + frame * res.xres *  res.yres*4); 
                  }            
 #endif
#endif		 



/*+*/		for (i=0; i<spes; i++) {
/*+*/			pthread_join (pts[i], NULL);
                  }    
                     
#ifndef WITHGRAPHICS
                calcitersum();  // wird sonst von draw übernommen
#endif
	
		if (fbok) {

			// false man synchronisieren möchte:
                        // wait for vsync interrupt 
			// uint32_t crt = 0; 
                        // ioctl(fd, FBIO_WAITFORVSYNC, (unsigned long)&crt);

#ifdef WITHGRAPHICS     // draw frame
 #ifndef USE2BUFFER
                draw1((uint32_t*) (u32) addr + frame * res.xres *  res.yres*4);
 #endif 
#endif


       //  umschalten auf zweiuten frame:     
       //  blit and flip with vsync request */
       //  ioctl(fd, PS3FB_IOCTL_FSEL, (unsigned long)&frame);

       //  switch frame
       //  frame = 1 - frame;
		}
                drx=drx/dfak;  
		di=drx;
   
       
      	}
#ifdef WITHGRAPHICS
 #ifdef USE2BUFFER 
	if (fbok) {
                  if (rep & 1) draw2 ((uint32_t*) (u32) addr + frame * res.xres *  res.yres*4);
                  else  draw1 ((uint32_t*) (u32) addr + frame * res.xres *  res.yres*4); 
                  }           
 #endif
#endif
       
  	te=seconds();
        gits=(double)gsum/1e9;
	printf("\nThe program has successfully executed.  %f GigaIterations \n",gits ); 
	//printf(" Maximal %f GigaIterations \n",(double)WIDTH*(double)HEIGHT*(double)Depth*rep/1e9); 

        printf ("r= %f, i= %f, depth= %d frames= %d, spes=  %d\n",rz,iz,Depth,Repmax,spes);
	printf("time: %fs => %f GIterations/s =>  %f GFlop/s\n",te-ta, gits/(te-ta), gits/(te-ta)*9); 

        if (!REPEATLOOP) break; 
        // für nächsten Durchgang bisschen zufällig herumhüpfen    
        rz=rz+(rand()%3-1)*hopping;
        iz=iz+(rand()%3-1)*hopping;

        z++;
     } while (!kbhit()); // bis Taste

/*+*/	for (i=0; i<spes; i++) {
/*+*/		spe_context_destroy(ctx[i]);
/*+*/	}
/*+*/

#ifdef WITHGRAPHICS       

        if (fbok){
		ioctl(fd, PS3FB_IOCTL_OFF, 0);
	
		munmap (addr,length); 
	}
	if (fd)  close (fd);

#endif

	return (0);
}
