/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <libspe2.h>
#include <pthread.h>
#include "spu/sgemm_spu.h"

extern spe_program_handle_t sgemm_spu;

#define SPU_THREADS 	6


double ta,te;
double flops;
struct timezone tz;

void *ppu_pthread_function(void *arg) {
  spe_context_ptr_t ctx;
  unsigned int entry = SPE_DEFAULT_ENTRY;
  
  ctx = *((spe_context_ptr_t *)arg);
  if (spe_context_run(ctx, &entry, 0, NULL, NULL, NULL) < 0) {
    perror ("Failed running context");
    exit (1);
  }
  pthread_exit(NULL);
}

double seconds(void) {
    struct timeval t;
    gettimeofday(&t,&tz);
    return (double) t.tv_sec+ ((double)t.tv_usec/1e6); 
}

int main()
{
  int i;
  spe_context_ptr_t ctxs[SPU_THREADS];
  pthread_t threads[SPU_THREADS];

  /* Create several SPE-threads to execute 'sgemm_spu'.
   */
  for(i=0; i<SPU_THREADS; i++) {
    /* Create context */
    if ((ctxs[i] = spe_context_create (0, NULL)) == NULL) {
      perror ("Failed creating context");
      exit (1);
    }
    /* Load program into context */
    if (spe_program_load (ctxs[i], &sgemm_spu)) {
      perror ("Failed loading program");
      exit (1);
    }
    ta=seconds();
    /* Create thread for each SPE context */
    if (pthread_create (&threads[i], NULL, &ppu_pthread_function, &ctxs[i]))  {
      perror ("Failed creating thread");
      exit (1);
    }
  }

  /* Wait for SPU-thread to complete execution.
   */
  for (i=0; i<SPU_THREADS; i++) {
    if (pthread_join (threads[i], NULL)) {
      perror("Failed pthread_join");
      exit (1);
    }
  }
  te=seconds();
  flops=SPU_THREADS*FLOPS; 
  printf("\nThe program has successfully executed after %f sec => %f GFlops/s.\n",te-ta,flops/(te-ta)/1e9 );

  return (0);
}
